/**
 * @file ixp425_eth.c
 *
 * @author Intel Corporation
 * @date 
 *
 * @brief IXP425 Ethernet Driver for Linux
 *
 * Design Notes:
 *
 * -- Intel Copyright Notice --
 *
 * @par
 * INTEL CONFIDENTIAL
 *
 * @par
 * Copyright 2003 Intel Corporation All Rights Reserved.
 *
 * @par
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel Corporation
 * or its suppliers or licensors.  Title to the Material remains with
 * Intel Corporation or its suppliers and licensors.  The Material
 * contains trade secrets and proprietary and confidential information of
 * Intel or its suppliers and licensors.  The Material is protected by
 * worldwide copyright and trade secret laws and treaty provisions. No
 * part of the Material may be used, copied, reproduced, modified,
 * published, uploaded, posted, transmitted, distributed, or disclosed in
 * any way without Intel's prior express written permission.
 *
 * @par
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise.  Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 *
 * @par
 * For further details, please see the file README.TXT distributed with
 * this software.
 * -- End Intel Copyright Notice --
 */


/* NOTE:
 * By default, this driver uses predefined MAC addresses.
 * These are set in global var 'default_mac_addr' in this file.
 * If required, these can be changed at run-time using
 * the 'ifconfig' tool.
 *
 * Example - to set ixp0 MAC address to 00:22:44:66:88:AA, 
 * run ifconfig with the following arguments:
 *   `ifconfig ixp0 hw ether 0022446688AA`
 */

/*
 * System-defined header files
 */
#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/delay.h>
#include <linux/mii.h>
#include <asm/io.h>
#include <asm/errno.h>

/*
 * IXP4xx CSR specific header files
 */
#include <ixp425.h>
#include <IxEthAcc.h>
#include <IxEthDB.h>
#include <IxQMgr.h>
#include <IxNpeDl.h>
#include <IxNpeMh.h>
#include <IxOsServices.h>
#include <IxOsBuffPoolMgt.h>
#include <IxEthNpe.h>
#include <IxNpeMicrocode.h>

/* memset buffer contents to CC [debug] */
#ifdef DEBUG
#define MEMSET_BUFF_CC
#endif


/* Hardcoded parameters */

/*
 * Module version information
 */
MODULE_DESCRIPTION("Ethernet driver for IXP425");
MODULE_LICENSE("Copyright (c) Intel Corporation. 2002");
MODULE_AUTHOR("Intel Corporation");
#define MODULE_NAME "ixp425_eth"
#define MODULE_VERSION "0.0.9"

/* Module parameters */
static int log_level = 0;
static int no_csr_init = 0;
static int no_phy_scan = 0;
MODULE_PARM(log_level, "i");
MODULE_PARM_DESC(log_level, "Set log level: 0 - None, 1 - Verbose, 2 - Debug");
MODULE_PARM(no_csr_init, "i");
MODULE_PARM_DESC(no_csr_init, "If non-zero, do not initialise CSR core components");
MODULE_PARM(no_phy_scan, "i");
MODULE_PARM_DESC(no_phy_scan, "If non-zero, use hard-coded phy addresses");

/* devices will be called ixp0 and ixp1 */
#define DEVICE_NAME "ixp"

/* boolean values for PHY link speed, duplex, and autonegotiation */
#define PHY_SPEED_10    0
#define PHY_SPEED_100   1
#define PHY_DUPLEX_HALF 0
#define PHY_DUPLEX_FULL 1
#define PHY_AUTONEG_OFF 0
#define PHY_AUTONEG_ON  1

/* Time before kernel will decide that the driver had stuck in transmit */
#define DEV_WATCHDOG_TIMEO (10*HZ)

/* Interval between media-sense/duplex checks */
#define MEDIA_CHECK_INTERVAL 3
/* Interval in milliseconds */
#define MEDIA_CHECK_INTERVAL_MS (MEDIA_CHECK_INTERVAL * 1000)

/* number of packets to prealloc for the Tx pool */
#define RX_QUEUE_PREALLOC 128

/* Maximum number of packets in Tx queue before netif_stop_queue() */
#define TX_QUEUE_MAX_LEN 128

/* Maximum number of addresses the EthAcc multicast filter can hold */
#define IX_ETH_ACC_MAX_MULTICAST_ADDRESSES 256

/* size of the SKBs to allocate */
#define PKT_SIZE 2048

/*
 * NPE s/w build ID. Use with IxNpeDl to select which microcode
 * firmware image to download to the NPEs.  The build ID indicates
 * the functionality of that image.  See also IxNpeDl.h.
 * The actual build Ids supported depends on the NPE and
 * the microcode image library used. 
 */
/* Functionality: Ethernet without FastPath, No Crypto */
#define IX_ETH_NPE_BUILD_ID_ETH                  0x00
/* Functionality: Ethernet with FastPath, No Crypto  */
#define IX_ETH_NPE_BUILD_ID_ETH_FPATH            0x01


/*
 * Macros to turn on/off debug messages
 */
/* Print kernel error */
#define P_ERROR(args...) \
    printk(KERN_ERR MODULE_NAME ": " args)
/* Print kernel warning */
#define P_WARN(args...) \
    printk(KERN_WARNING MODULE_NAME ": " args)
/* Print kernel notice */
#define P_NOTICE(args...) \
    printk(KERN_NOTICE MODULE_NAME ": " args)
/* Print kernel info */
#define P_INFO(args...) \
    printk(KERN_INFO MODULE_NAME ": " args)
/* Print verbose message. Enabled/disabled by 'log_level' param */
#define P_VERBOSE(args...) \
    if (log_level >= 1) printk(MODULE_NAME ": " args)
/* Print debug message. Enabled/disabled by 'log_level' param  */
#define P_DEBUG(args...) \
    if (log_level >= 2) printk(MODULE_NAME ": " __FUNCTION__ "(): " args)
/* Print trace message */
#define TRACE P_DEBUG("line %d\n", __LINE__)


/* Private device data */
typedef struct {
    spinlock_t lock;

    IxEthAccPortId port_id; /* MAC port_id */

    /* id of thread for the link duplex monitoring */
    int mediaCheckThreadId;

    /* mutex locked by thread, until the thread exits */
    struct semaphore *mediaCheckThreadComplete;

    struct net_device_stats stats; /* device statistics */

    /* preallocated RX pool */
    IX_MBUF rx_pool_buf[RX_QUEUE_PREALLOC];
    IX_MBUF_POOL *rx_pool;

    /* preallocated TX pool */
    IX_MBUF tx_pool_buf[TX_QUEUE_MAX_LEN];
    IX_MBUF_POOL *tx_pool;

    /* Set to 1 in dev->open and to 0 in dev->stop.
     * Checked by rx_cb to decide if to pass packet to the kernel [1] or
     * back to the pool [0].
     * Also used to stop the kernel thread for link monitoring.
     */
    int running;

} priv_data_t;

/* Collection of boolean PHY configuration parameters */
typedef struct {
    int speed100;
    int duplexFull;
    int autoNegEnabled;
} phy_cfg_t;
	

/*
 * STATIC VARIABLES
 *
 * This section sets several default values for each port.
 * These may be edited if required.
 */


/* The PHY addresses mapped to CSR EthAcc ports.
 *
 * These are hardcoded here but typically overwritten
 * by a PHY scan. A PHY scan is enabled by default but
 * can optionally be disabled if required in 2 ways:
 *  - by setting CONFIG_IXP425_CSR_ETH_SKIP_PHY_SCAN
 *  - by passing module param no_phy_scan with non-zero value
 *
 * Up to 32 PHYs may be discovered by a phy scan. Addresses
 * of all PHYs found will be stored here, but only the first
 * 2 will be used with the CSR EthAcc ports.
 *
 * See also the function phy_init() in this file.
 *
 * NOTE: The hardcoded PHY addresses have been verified on
 * the IXDP425 and Coyote (IXP4XX RG) Development platforms.
 * However, they may differ on other platforms.
 */
static int phyAddresses[IXP425_ETH_ACC_MII_MAX_ADDR] =
{
#ifdef CONFIG_ARCH_IXDP425
    0, /* Port 1 (ixp0) */
    1  /* Port 2 (ixp1) */

#elif CONFIG_ARCH_IXP425_COYOTE
    4, /* Port 1 (ixp0) - Connected to PHYs 1-4               */
    5, /* Port 2 (ixp1) - Only connected to PHY 5             */

    3,  /******************************************************/
    2,  /* PHY addresses on Coyote platform (physical layout) */
    1   /* (4 LAN ports, switch)  (1 WAN port)                */
        /*       ixp0              ixp1                       */
        /*  ________________       ____                       */
        /* /_______________/|     /___/|                      */
	/* | 1 | 2 | 3 | 4 |      | 5 |                       */
        /* ----------------------------------------           */

#else
    0, /* Port 1 (ixp0) */
    1  /* Port 2 (ixp1) */

#endif
};

/* The default configuration of the phy on each CSR EthAcc port.
 *
 * This configuration is loaded when the phy's are discovered.
 * More PHYs can optionally be configured here by adding more
 * configuration entries in the array below.  The PHYs will be
 * configured in the order that they are found, starting with the
 * lowest PHY address.
 *
 * See also function phy_init() in this file
 */
static phy_cfg_t default_phy_cfg[] =
{
#ifdef CONFIG_ARCH_IXDP425
    {PHY_SPEED_100, PHY_DUPLEX_FULL, PHY_AUTONEG_ON}, /* Port 1 */
    {PHY_SPEED_100, PHY_DUPLEX_FULL, PHY_AUTONEG_ON}  /* Port 2 */

#elif CONFIG_ARCH_IXP425_COYOTE
    {PHY_SPEED_100, PHY_DUPLEX_FULL, PHY_AUTONEG_ON}, /* Port 1 */
    {PHY_SPEED_100, PHY_DUPLEX_FULL, PHY_AUTONEG_ON}, /* Port 2 */
    {PHY_SPEED_100, PHY_DUPLEX_FULL, PHY_AUTONEG_ON},
    {PHY_SPEED_100, PHY_DUPLEX_FULL, PHY_AUTONEG_ON},
    {PHY_SPEED_100, PHY_DUPLEX_FULL, PHY_AUTONEG_ON}

#else
    {PHY_SPEED_100, PHY_DUPLEX_FULL, PHY_AUTONEG_ON}, /* Port 1 */
    {PHY_SPEED_100, PHY_DUPLEX_FULL, PHY_AUTONEG_ON}  /* Port 2 */

#endif
};

/* Default MAC addresses for ixp0 and ixp1 (using Intel MAC prefix) */
static IxEthAccMacAddr default_mac_addr[IX_ETH_ACC_NUMBER_OF_PORTS] =
{
    {{0x00, 0x02, 0xB3, 0x01, 0x01, 0x01}}, /* Port 1 */
    {{0x00, 0x02, 0xB3, 0x02, 0x02, 0x02}}  /* Port 2 */
};

/* NPE firmware build Ids.  These are used to select the
 * correct firmware image to download to each NPE.
 * The build ID indicates the core functionality of the image.
 */
static IxNpeDlBuildId default_npe_buildId[IX_ETH_ACC_NUMBER_OF_PORTS] =
{
    IX_ETH_NPE_BUILD_ID_ETH_FPATH,  /* Port 1 */
    IX_ETH_NPE_BUILD_ID_ETH         /* Port 2 */
};

/* Mutex lock used to coordinate access to IxEthAcc functions
 * which manipulate the MII registers on the PHYs
 */
static struct semaphore *miiAccessMutex;


/*
 * CSR ERROR NUMBER FUNCTIONS
 */
/* Convert IxEthAcc return codes to suitable Linux return codes */
int convert_error_ethAcc (IxEthAccStatus error)
{
    switch (error)
    {
	case IX_ETH_ACC_SUCCESS:            return  0;
	case IX_ETH_ACC_FAIL:               return -1;
	case IX_ETH_ACC_INVALID_PORT:       return -ENODEV;
	case IX_ETH_ACC_PORT_UNINITIALIZED: return -EPERM;
	case IX_ETH_ACC_MAC_UNINITIALIZED:  return -EPERM;
	case IX_ETH_ACC_INVALID_ARG:        return -EINVAL;
	case IX_ETH_TX_Q_FULL:              return -EAGAIN;
	case IX_ETH_ACC_NO_SUCH_ADDR:       return -EFAULT;
	default:                            return -1;
    };
}


/*
 * DEBUG UTILITY FUNCTIONS
 */
#ifdef DEBUG_DUMP

static void hex_dump(void *buf, int len)
{
    int i;

    for (i = 0 ; i < len; i++)
    {
	printk("%02x", ((u8*)buf)[i]);
	if (i%2)
	    printk(" ");
	if (15 == i%16)
	    printk("\n");
    }
    printk("\n");
}

static void mbuf_dump(char *name, IX_MBUF *mbuf)
{
    printk("+++++++++++++++++++++++++++\n"
	"%s MBUF dump mbuf=%p, m_data=%p, m_len=%d, len=%d\n",
	name, mbuf, mbuf->m_data, mbuf->m_len, mbuf->m_pkthdr.len);
    printk(">> mbuf:\n");
    hex_dump(mbuf, sizeof(*mbuf));
    printk(">> m_data:\n");
    hex_dump(__va(mbuf->m_data), mbuf->m_pkthdr.len);
    printk("\n-------------------------\n");
}

static void skb_dump(char *name, struct sk_buff *skb)
{
    printk("+++++++++++++++++++++++++++\n"
	"%s SKB dump skb=%p, data=%p, len=%d\n",
	name, skb, skb->data, skb->len);
    printk(">> data:\n");
    hex_dump(skb->data, skb->len);
    printk("\n-------------------------\n");
}
#endif


/*
 * BUFFER MANAGEMENT
 */

/* Return mbuf+skb pair to the CSR.
 * In the case of an error, free skb and return mbuf to the pool
 */
static int rx_buff_replenish(int port_id, IX_MBUF *mbuf)
{
    int res;

    if ((res = ixEthAccPortRxFreeReplenish(port_id, mbuf)))
    {
	P_WARN("ixEthAccPortRxFreeReplenish failed for port %d, res = %d",
		port_id, res);

	dev_kfree_skb_any(mbuf_swap_skb(mbuf, NULL));
	IX_MBUF_POOL_PUT(mbuf);

	return convert_error_ethAcc(res);
    }

    return 0;
}

/* Allocate new skb and replenish the pair.
 * in the case of an error, mbuf will be returned to the pool.
 * Note: in any case, [old] skb will be extracted from the mbuf and
 * returned in skbret(if not NULL)
 */
static int rx_buff_recycle(int port_id, IX_MBUF *mbuf, struct sk_buff **skbret)
{
    struct sk_buff *skb = dev_alloc_skb(PKT_SIZE);

    /* eth header is 14 bytes, and we need ip header to be alligned
     * by 16 bytes
     */
    if (skb)
	skb_reserve(skb, 2);

    /* Our USB driver needs 48 byte reserved to avoid memcpy */
#ifdef CONFIG_USB_RNDIS
    skb_reserve(skb, 48);
#endif

    /* extract skb from mbuf */
    if (skbret)
	*skbret = mbuf_swap_skb(mbuf, skb);
    else
	mbuf_swap_skb(mbuf, skb);

    if (!skb)
    {
	/* failed to alloc skb -> return mbuf to the pool, it'll be
	 * picked up later
	 */
	IX_MBUF_POOL_PUT(mbuf);
	return -ENOMEM;
    }

#ifdef MEMSET_BUFF_CC
    memset(skb->head, 0xCC, skb->end - skb->head);
#endif

    return rx_buff_replenish(port_id, mbuf);
}

/* Allocate an skb for every mbuf in the rx_pool
 * and pass the pair to the CSR
 */
static int rx_buff_prealloc(struct net_device *dev)
{
    int res;
    priv_data_t *priv = dev->priv;
    IX_MBUF *mbuf;

    while ((IX_MBUF_POOL_GET(priv->rx_pool, &mbuf) == IX_SUCCESS))
    {
	/* we pass NULL as skbret, becouse an mbuf from the pool
	 * does not have an skb attached
	 */
	if ((res = rx_buff_recycle(priv->port_id, mbuf, NULL)))
	    return res;
    }

    return 0;
}


/* 
 * KERNEL THREADS
 */

/* This timer will check the PHY for the link duplex and
 * update the MAC accordingly
 */
static int
media_check_thread (void* arg)
{
    struct net_device *dev = (struct net_device *) arg;
    priv_data_t *priv = dev->priv;
    int linkUp;
    int speed100;
    int fullDuplex;
    int autonegotiate;
    unsigned phyNum = phyAddresses[priv->port_id];
    int res;

    TRACE;

    /* Lock the mutex for this thread.
       This mutex can be used to wait until the thread exits
    */
    down (priv->mediaCheckThreadComplete);

    daemonize();
    reparent_to_init();
    spin_lock_irq(&current->sigmask_lock);
    sigemptyset(&current->blocked);
    recalc_sigpending(current);
    spin_unlock_irq(&current->sigmask_lock);
    
    snprintf(current->comm, sizeof(current->comm), "ixp425_eth %s", dev->name);

    TRACE;
    
    while (1)
    {
	/* We may have been woken up by a signal. If so, we need to flush it out */
	if (signal_pending (current)) {
	    spin_lock_irq(&current->sigmask_lock);
	    flush_signals(current);
	    spin_unlock_irq(&current->sigmask_lock);
	}

	/* If the interface is down, we need to exit this thread */
	if (!priv->running)
	{
	    break;
	}

	/*
	 * Determine the link status
	 */

	TRACE;

	/* lock the MII register access mutex */
	down(miiAccessMutex);
	res = ixEthAccMiiLinkStatus(phyNum,
				    &linkUp,
				    &speed100,
				    &fullDuplex, 
				    &autonegotiate);
	/* release the MII register access mutex */
	up(miiAccessMutex);

	if (res != IX_ETH_ACC_SUCCESS)
	{
	    P_WARN("ixEthAccMiiLinkStatus failed on PHY%d.\n"
		   "\tCan't determine\nthe auto negotiated parameters. "
		   "Using default values.\n",
		   phyNum); 

	    /* this shouldn't happen. exit the thread if it does */
	    break;
	}
    
	if (linkUp)
	{
	    if (! netif_carrier_ok(dev))
	    {
		/* inform the kernel of a change in link state */
		netif_carrier_on(dev);
	    }

	    /*
	     * Update the MAC mode to match the PHY mode
	     */
	    if (fullDuplex)
	    {
		ixEthAccPortDuplexModeSet (priv->port_id, IX_ETH_ACC_FULL_DUPLEX);
	    }
	    else
	    {
		ixEthAccPortDuplexModeSet (priv->port_id, IX_ETH_ACC_HALF_DUPLEX);
	    }
	}
	else
	{
	    if (netif_carrier_ok(dev))
	    {
		/* inform the kernel of a change in link state */
		netif_carrier_off(dev);
	    }
	}
    
	TRACE;
    
	/* this is to prevent CSR rx pool from emptying when
	 * there's not enough memory for a long time
	 */
	rx_buff_prealloc(dev);
    
	/* Now sleep for 3 seconds */
	current->state = TASK_INTERRUPTIBLE;
	schedule_timeout((MEDIA_CHECK_INTERVAL_MS * HZ) / 1000);
    } /* while (1) ... */

    /* free the mutex for this thread. */
    up (priv->mediaCheckThreadComplete);

    return 0;
}


/*
 * TIMERS
 */

/* This timer will call ixEthDBDatabaseMaintenance every
 * IX_ETH_DB_MAINTENANCE_TIME seconds
 */
static void maintenance_timer_cb(unsigned long data);

static struct timer_list maintenance_timer = {
    function:&maintenance_timer_cb
};

static void maintenance_timer_set(void)
{
    maintenance_timer.expires = jiffies + IX_ETH_DB_MAINTENANCE_TIME * HZ;
    add_timer(&maintenance_timer);
}

static void maintenance_timer_clear(void)
{
    del_timer_sync(&maintenance_timer);
}

static void maintenance_timer_cb(unsigned long data)
{
    ixEthDBDatabaseMaintenance();

    maintenance_timer_set();
}

/* This callback is called when transmission of the packed is done, and
 * IxEthAcc does not need the buffer anymore.
 */
static void tx_done_cb(u32 callbackTag, IX_MBUF *mbuf)
{
    struct net_device *dev = (struct net_device *)callbackTag;
    priv_data_t *priv = dev->priv;

    TRACE;

    priv->stats.tx_packets++; /* total packets transmitted */
    priv->stats.tx_bytes += mbuf->m_len; /* total bytes transmitted */

    /* extract skb from the mbuf, free skb and return the mbuf to the pool */
    dev_kfree_skb_any(mbuf_swap_skb(mbuf, NULL));
    IX_MBUF_POOL_PUT(mbuf);

    /* Start the queue in case it was stopped in dev_hard_start_xmit
     * (when queue-full detected)
     */
    if (priv->running && netif_queue_stopped(dev))
    {
	TRACE;
	netif_wake_queue(dev);
    }
}

/* This callback is called when new packet received from MAC
 * and ready to be transfered up-stack.
 *
 * If this is a valid packet, then new skb is allocated and switched
 * with the one in mbuf, which is pushed upstack.
 *
 * In the case of en empty/error packet, the mbuf just pushed
 * back to the CSR [unless device is down, in which case it's returned to
 * the pool].
 */
static void rx_cb(u32 callbackTag, IX_MBUF *mbuf, IxEthAccPortId portId)
{
    struct net_device *dev = (struct net_device *)callbackTag;
    priv_data_t *priv = dev->priv;
    struct sk_buff *skb;
    int len;

    if (mbuf->m_len)
    {
	TRACE;

	/* valid packet */

	len = mbuf->m_len; /* preserve the mbuf len */

	/* allocate new skb and "swap" it with the skb that is tied to the
	 * mbuf. then return the mbuf + new skb to the CSR.
	 */
	rx_buff_recycle(priv->port_id, mbuf, &skb);

	skb_put(skb, len);
	skb->dev = dev;
	skb->protocol = eth_type_trans(skb, dev);

	/* push upstack */
	TRACE;

	priv->stats.rx_packets++; /* total packets received */
	priv->stats.rx_bytes += len; /* total bytes received */

	netif_rx(skb);
    }
    else if (!priv->running)
    {
	TRACE;

	/* this is a "junk" buffer, returned by the
	 * call to PortDisable
	 */

	/* free the skb & return the mbuf to the pool */
	dev_kfree_skb_any(mbuf_swap_skb(mbuf, NULL));
	IX_MBUF_POOL_PUT(mbuf);

	return;
    }
    else
    {
	TRACE;
	/* Should not get here!
	 * error buf!!! -> push the mbuf back to the csr */
	rx_buff_replenish(priv->port_id, mbuf);
    }

    TRACE;
}

/* Set promiscuous/multicast mode for the MAC */
static void dev_set_multicast_list(struct net_device *dev)
{
    int res;
    priv_data_t *priv = dev->priv;
    IxEthAccMacAddr addr1 = {};

/* 4 possible scenarios here
 *
 * scenarios:
 * #1 - promiscuous mode ON
 * #2 - promiscuous mode OFF, accept NO multicast addresses
 * #3 - promiscuous mode OFF, accept ALL multicast addresses
 * #4 - promiscuous mode OFF, accept LIST of multicast addresses 
 */

    TRACE;

    /* if called from irq handler, lock already acquired */
    if (!in_irq())
	spin_lock_irq(&priv->lock);

    /* clear multicast addresses that were set the last time (if exist) */
    ixEthAccPortMulticastAddressLeaveAll (priv->port_id);

/**** SCENARIO #1 ****/
    /* Set promiscuous mode */
    if (dev->flags & IFF_PROMISC)
    {
	if ((res = ixEthAccPortPromiscuousModeSet(priv->port_id)))
	{
	    P_ERROR("%s: ixEthAccPortPromiscuousModeSet failed on port %d\n",
		    dev->name, priv->port_id);
	}
	else
	{
	    P_NOTICE("%s: Entering promiscuous mode\n", dev->name);
	}

	goto Exit;
    }


/**** SCENARIO #2 ****/

    /* Clear promiscuous mode */
    if ((res = ixEthAccPortPromiscuousModeClear(priv->port_id)))
    {
	/* should not get here */
	P_ERROR("%s: ixEthAccPortPromiscuousModeClear failed for port %d\n",
		dev->name, priv->port_id);
    }
    else
    {
	P_NOTICE("%s: Leaving promiscuous mode\n", dev->name);
    }


/**** SCENARIO #3 ****/
    /* If there's more addresses than we can handle, get all multicast
     * packets and sort the out in software
     */
    /* Set multicast mode */
    if ((dev->flags & IFF_ALLMULTI) || 
	(dev->mc_count > IX_ETH_ACC_MAX_MULTICAST_ADDRESSES))
    {
	/* Get ALL multicasts:
	 * This will result in
         *     MASK == 01:00:00:00:00:00
         *     ADDR == 01:00:00:00:00:00
	 * Meaning ALL MULTICAST addresses will be accepted
	 */
        ixEthAccPortMulticastAddressJoinAll(priv->port_id);

	P_VERBOSE("%s: Accepting ALL multicast packets\n", dev->name);
	goto Exit;
    }


/**** SCENARIO #4 ****/
    /* Store all of the multicast addresses in the hardware filter */
    if ((dev->mc_count))
    {
	/* now join the current address list */
	/* Get only multicasts from the list */
	struct dev_mc_list *mc_ptr;

	for (mc_ptr = dev->mc_list; mc_ptr; mc_ptr = mc_ptr->next)
	{
	    memcpy (&addr1.macAddress[0], &mc_ptr->dmi_addr[0],
		    IX_IEEE803_MAC_ADDRESS_SIZE);
	    ixEthAccPortMulticastAddressJoin (priv->port_id, &addr1);
	}
    }


Exit:
    if (!in_irq())
	spin_unlock_irq(&priv->lock);
}


static int npe_init(IxNpeDlNpeId npeId, IxNpeDlBuildId buildId)
{
    int res;
    IxNpeDlVersionId npeIdList[IX_NPE_MICROCODE_AVAILABLE_VERSIONS_COUNT];
    u32 npeListSize = IX_NPE_MICROCODE_AVAILABLE_VERSIONS_COUNT;

    IxNpeDlVersionId tmpIdList = {npeId, buildId, 0, 0};
    int i;

    /* Stop the NPE before attempting a download */
    if (ixNpeDlNpeStopAndReset(npeId) != IX_SUCCESS)
    {
	P_ERROR("NPE stop and reset failed\n");
	return -1;
    }

    /* Get a list of all the NPE firmware images available */
    if ((res = ixNpeDlAvailableVersionsListGet(npeIdList, &npeListSize)))
    {
	P_ERROR("Unable to get NPE version information\n");
	return -1;
    }

    /* search the list for the newest image for this NPE 
     * with the specified functionality
     */
    for (i = 0; i < npeListSize; i++)
    {
        if ((npeIdList[i].npeId == npeId) && (npeIdList[i].buildId == buildId))
	{
	    if (npeIdList[i].major > tmpIdList.major)
	    {
		tmpIdList.major = npeIdList[i].major;
		tmpIdList.minor = npeIdList[i].minor;
	    }
	    else if (npeIdList[i].major == tmpIdList.major)
	    {
		if (npeIdList[i].minor >= tmpIdList.minor)
		{
		    tmpIdList.minor = npeIdList[i].minor;
		}
	    }
	}
    }

    P_VERBOSE("\nNPE %d Version information\n", tmpIdList.npeId);
    P_VERBOSE("Major ID:\t 0x%x\n", tmpIdList.major);
    P_VERBOSE("Minor ID:\t 0x%x\n", tmpIdList.minor);
    P_VERBOSE("Build ID:\t 0x%x\n", tmpIdList.buildId);

    /* Download NPE code */
    if ((res = ixNpeDlVersionDownload(&tmpIdList, TRUE)))
    {
	P_ERROR("NPE download failed\n");
	return -1;
    }

    return 0;
}

static int npe_start(IxEthAccPortId port_id)
{
    int res;
    IxNpeDlNpeId npeId;
    IxNpeDlBuildId buildId;

    switch (port_id)
    {
        case IX_ETH_PORT_1:
	    npeId = IX_NPEDL_NPEID_NPEB;
	    break;
        case IX_ETH_PORT_2:
	    npeId = IX_NPEDL_NPEID_NPEC;
	    break;
	default:
	    P_ERROR("Invalid port specified. IXP Ethernet NPE not started\n");
	    return -ENODEV;
    }
    buildId = default_npe_buildId[port_id];

    /* Initialise NPE */
    if ((res = npe_init(npeId, buildId)))
    {
	P_ERROR("Error initialising NPE for Ethernet port %d!\n", port_id);
	return res;
    }

    /* Start NPEs */
    if ((res = ixNpeDlNpeExecutionStart(npeId)))
    {
	P_ERROR("Error starting NPE for Ethernet port %d!\n", port_id);
	return -1;
    }

    return 0;
}

/* Enable the MAC port.
 * Called on do_dev_open and dev_tx_timeout
 */
static int port_enable(struct net_device *dev)
{
    int res;
    priv_data_t *priv = dev->priv;

    P_DEBUG("port_enable(%s)\n", dev->name);

    /* we need to restart the npe's here in case they have been
     * stopped by port_disable()
     */
    if (!priv->running)
    {
	if ((res = npe_start(priv->port_id)))
	{
	    TRACE;
	    return res;
	}
    }

    if ((res = ixEthAccPortTxFrameAppendFCSEnable(priv->port_id)))
    {
        TRACE;
        return convert_error_ethAcc(res);
    }

    if ((res = ixEthAccPortRxFrameAppendFCSDisable(priv->port_id)))
    {
	TRACE;
	return convert_error_ethAcc(res);
    }

    if((res = ixEthAccTxSchedulingDisciplineSet(priv->port_id,
						FIFO_NO_PRIORITY)))
    {
        TRACE;
	return convert_error_ethAcc(res);
    }

    if ((res = ixEthAccPortEnable(priv->port_id)))
    {
	P_ERROR("%s: ixEthAccPortEnable failed for port %d, res = %d\n",
		dev->name, priv->port_id, res);
	return convert_error_ethAcc(res);
    }

    TRACE;

    priv->running = 1;

    netif_start_queue(dev);

    TRACE;

    return 0;
}

/* Disable the MAC port.
 * Called on do_dev_stop and dev_tx_timeout
 */
static void port_disable(struct net_device *dev)
{
    priv_data_t *priv = dev->priv;
    int res;

    TRACE;

    netif_stop_queue(dev);

    priv->running = 0;

    /* Wake up the media-check thread with a signal.
       It will check the 'running' flag and exit */
    if ((res = kill_proc (priv->mediaCheckThreadId, SIGKILL, 1)))
    {
	P_ERROR("%s: unable to signal thread\n", dev->name);
    }
    else
    {
	/* wait for the thread to exit. */
	down (priv->mediaCheckThreadComplete);
	up (priv->mediaCheckThreadComplete);
    }

    if ((res = ixEthAccPortDisable(priv->port_id)))
    {
	/* should not get here */
	P_ERROR("%s: BUG: ixEthAccPortDisable(%d) failed with res=%d\n",
		dev->name, priv->port_id, res);
    }

    TRACE;
}

/* Open the device.
 * Request resources and start interrupts
 */
static int do_dev_open(struct net_device *dev)
{
    int res;
    priv_data_t *priv = dev->priv;

    /* Preallocate RX pool */
    P_DEBUG("%s: Preallocating buffers\n", dev->name);
    if ((res = rx_buff_prealloc(dev)))
	return res;

    /* Set callbacks */
    if ((res = ixEthAccPortTxDoneCallbackRegister(priv->port_id, tx_done_cb,
	(u32)dev)))
    {
	P_ERROR("%s: ixEthAccPortTxDoneCallbackRegister for port %d failed, "
		"res = %d\n", dev->name, priv->port_id, res);
	return convert_error_ethAcc(res);
    }

    if ((res = ixEthAccPortRxCallbackRegister(priv->port_id, rx_cb,
	(u32)dev)))
    {
	P_ERROR("%s: ixEthAccPortRxCallbackRegister for port %d failed, "
		"res = %d\n", dev->name, priv->port_id, res);
	return convert_error_ethAcc(res);
    }

    if ((res = port_enable(dev)))
    {
        return res;
    }
    
    /* Invoke the link duplex monitor */
    priv->mediaCheckThreadId = kernel_thread(media_check_thread,
					     (void *) dev,
					     CLONE_FS | CLONE_FILES);
    if (priv->mediaCheckThreadId < 0)
    {
	P_ERROR("%s: Failed to start thread for media checks\n", dev->name);
    }

    return res;
}

/* Close the device.
 * Free resources acquired in dev_start
 */
static int do_dev_stop(struct net_device *dev)
{
    TRACE;

    netif_stop_queue(dev);

    port_disable(dev);

    return 0;
}

/* this is called by kernel to transmit packet */
static int dev_hard_start_xmit(struct sk_buff *skb, struct net_device *dev)
{
    int res;
    priv_data_t *priv = dev->priv;
    IX_MBUF *mbuf;

    TRACE;

    /* get mbuf struct from tx_pool */
    IX_MBUF_POOL_GET(priv->tx_pool, &mbuf);
    if (!mbuf)
	BUG(); /* Shouldn't happen */
    
    /* if pool has no more buffers */
    if (IX_MBUF_POOL_FREE_COUNT(priv->tx_pool) == 0)
    {
	TRACE;
	netif_stop_queue(dev);
    }

    mbuf_swap_skb(mbuf, skb); /* this should return NULL, as mbufs in the pool
    			       * have no skb attached
			       */

    /* flush the mbuf data from the cache */
    IX_ACC_DATA_CACHE_FLUSH(IX_MBUF_MDATA(mbuf), IX_MBUF_MLEN(mbuf));

    if ((res = ixEthAccPortTxFrameSubmit(priv->port_id, mbuf,
	IX_ETH_ACC_TX_DEFAULT_PRIORITY)))
    {
	P_ERROR("%s: ixEthAccPortTxFrameSubmit failed for port %d, res = %d\n",
		dev->name, priv->port_id, res);
	netif_stop_queue(dev);
	return convert_error_ethAcc(res);
    }

    return 0;
}

/* This function is called when kernel thinks that TX is stuck */
static void dev_tx_timeout(struct net_device *dev)
{
    TRACE;

    /* just restart the port */
    port_disable(dev);
    port_enable(dev);
}

static int dev_do_ioctl(struct net_device *dev, struct ifreq *req, int cmd)
{
    priv_data_t *priv = dev->priv;
    struct mii_ioctl_data *data = (struct mii_ioctl_data *) & req->ifr_data;
    int phy = phyAddresses[priv->port_id];
    int res = 0;

    TRACE;

    switch (cmd)
    {
	/*
	 * IOCTL's for mii-tool support
	 */

	/* Get address of MII PHY in use */
	case SIOCGMIIPHY:
	case SIOCDEVPRIVATE:
	    data->phy_id = phy;

        /* Read MII PHY register */
	case SIOCGMIIREG:		
	case SIOCDEVPRIVATE+1:
	    down (miiAccessMutex);     /* lock the MII register access mutex */
	    if ((res = ixEthAccMiiReadRtn (data->phy_id, data->reg_num, &data->val_out)))
	    {
		P_ERROR("Error reading MII reg %d on phy %d\n",
		       data->phy_id, data->reg_num);
		res = -1;
	    }
	    up (miiAccessMutex);	/* release the MII register access mutex */
	    return res;

	/* Write MII PHY register */
	case SIOCSMIIREG:
	case SIOCDEVPRIVATE+2:
	    down (miiAccessMutex);     /* lock the MII register access mutex */
	    if ((res = ixEthAccMiiWriteRtn (data->phy_id, data->reg_num, data->val_in)))
	    {
		P_ERROR("Error reading MII reg %d on phy %d\n",
			data->phy_id, data->reg_num);
		res = -1;
	    }
	    up (miiAccessMutex);	/* release the MII register access mutex */
	    return res;

	default:
	    return -EOPNOTSUPP;
    }

    return -ENODEV;
}

static struct net_device_stats *dev_get_stats(struct net_device *dev)
{
    int res;
    /* "stats" should be cache-safe.
     * we alligne "stats" and "priv" by 32 bytes, so that the cache
     * operations will not affect "res" and "priv"
     */
    IxEthEthObjStats stats __attribute__ ((aligned(32))) = {};
    priv_data_t *priv __attribute__ ((aligned(32)));

    TRACE;

    priv = dev->priv;

    /* Get HW stats and translate to the net_device_stats */
    if (!netif_running(dev))
    {
	TRACE;
	return &priv->stats;
    }

    TRACE;

    clean_dcache_range((u32)&stats, (u32)((&stats) + 1) - 1);
    if ((res = ixEthAccMibIIStatsGetClear(priv->port_id, &stats)))
    {
	P_ERROR("%s: ixEthAccMibIIStatsGet failed for port %d, res = %d\n",
		dev->name, priv->port_id, res);
	return &priv->stats;
    }

    TRACE;

    invalidate_dcache_range((u32)&stats, (u32)((&stats) + 1) - 1);

    /* bad packets received */
    priv->stats.rx_errors += 
	stats.dot3StatsAlignmentErrors +
	stats.dot3StatsFCSErrors +
	stats.dot3StatsFrameTooLongs +
	stats.dot3StatsInternalMacReceiveErrors;

    /* packet transmit problems */
    priv->stats.tx_errors += 
	stats.dot3StatsLateCollisions +
	stats.dot3StatsExcessiveCollsions +
	stats.dot3StatsInternalMacTransmitErrors +
	stats.dot3StatsCarrierSenseErrors;

    priv->stats.collisions +=
	stats.dot3StatsSingleCollisionFrames +
	stats.dot3StatsMultipleCollisionFrames;

    /* recved pkt with crc error */
    priv->stats.rx_crc_errors +=
	stats.dot3StatsFCSErrors;

    /* recv'd frame alignment error */
    priv->stats.rx_frame_errors +=
	stats.dot3StatsFrameTooLongs;

    /* detailed tx_errors */
    priv->stats.tx_carrier_errors +=
	stats.dot3StatsCarrierSenseErrors;

    return &priv->stats;
}

/* Initialize QMgr and bint it's interrupts */
static int qmgr_init(void)
{
    int res;

    /* Initialise Queue Manager */
    P_VERBOSE("Initialising Queue Manager...\n");
    if ((res = ixQMgrInit()))
    {
	P_ERROR("Error initialising queue manager!\n");
	return -1;
    }

    /*
     * Hook the QM QLOW dispatcher to the interrupt controller.
     * The ethernet NPEs use queues 24 through 27.
     */
    if ((res = ixOsServIntBind(IXP425_INT_LVL_QM1,
	(void(*)(void*))ixQMgrDispatcherLoopRun, (void *)IX_QMGR_QUELOW_GROUP)))
    {
	P_ERROR("Failed to bind to QM1 interrupt!\n");
	return -1;
    }

    return 0;
}

static int ethacc_init(void)
{
    int res;

    if ((res = ixEthAccInit()))
    {
	P_ERROR("ixEthAccInit failed with res=%d\n", res);
	return convert_error_ethAcc(res);
    }

    if ((res = ixEthAccPortInit(IX_ETH_PORT_1)))
    {
	P_ERROR("Failed to initialize Eth port 1 res = %d\n", res);
	return convert_error_ethAcc(res);
    }

    if ((res = ixEthAccPortInit(IX_ETH_PORT_2)))
    {
	P_ERROR("Failed to initialize Eth port 2 res = %d\n", res);
	return convert_error_ethAcc(res);
    }

    return 0;
}

#ifndef CONFIG_IXP425_CSR_ETH_SKIP_PHY_SCAN

static int phy_init(void)
{
    int res;
    int physcan[IXP425_ETH_ACC_MII_MAX_ADDR];
    int i, phy_found, num_phys_to_set;

    /* Use hardcoded phy addresses */
    if (no_phy_scan) /* Module parameter */
    {
	num_phys_to_set = (sizeof(default_phy_cfg) / sizeof(phy_cfg_t));
    }
    /* Do a PHY scan */
    else
    {
	if ((res = ixEthAccMiiPhyScan(physcan)))
	{
	    P_ERROR("PHY scan failed\n");
	    return convert_error_ethAcc(res);
	}
	
	for (i=0, phy_found=0; i < IXP425_ETH_ACC_MII_MAX_ADDR; i++)
	{
	    if (physcan[i])
	    {
		P_INFO("Found PHY %d at %d\n", phy_found, i);
		phyAddresses[phy_found] = i;
		
		if (++phy_found == IXP425_ETH_ACC_MII_MAX_ADDR)
		    break;
	    }
	}

	num_phys_to_set = phy_found;
    }

    /* Reset and Set each phy properties */
    for (i=0; i < num_phys_to_set; i++) 
    {
	P_VERBOSE("Configuring PHY %d\n", i);
	P_VERBOSE("\tSpeed %s\tDuplex %s\tAutonegotiation %s\n",
		  (default_phy_cfg[i].speed100) ? "100" : "10",   
		  (default_phy_cfg[i].duplexFull) ? "FULL" : "HALF",  
		  (default_phy_cfg[i].autoNegEnabled) ? "ON" : "OFF");

	ixEthAccMiiPhyReset(phyAddresses[i]);

	ixEthAccMiiPhyConfig(phyAddresses[i],
	    default_phy_cfg[i].speed100,   
	    default_phy_cfg[i].duplexFull,  
	    default_phy_cfg[i].autoNegEnabled);
    }

    for (i=0; i < IX_ETH_ACC_NUMBER_OF_PORTS; i++) 
    {
	P_INFO("%s%d is using the PHY at address %d\n",
	       DEVICE_NAME, i, phyAddresses[i]);

	/* Set the MAC to the same duplex mode as the phy */
	ixEthAccPortDuplexModeSet(i,
            (default_phy_cfg[i].duplexFull) ?
                 IX_ETH_ACC_FULL_DUPLEX : IX_ETH_ACC_HALF_DUPLEX);
    }

    return 0;
}
#endif

/* This function set's default MAC addresses for the ports.
 * They will be used until dev->set_mac_address hook is called
 */
static int set_def_mac_addr(IxEthAccPortId port)
{
    int res;

    if ((res = ixEthAccPortUnicastMacAddressSet(port, &default_mac_addr[port])))
    {
	P_ERROR("Failed to set MAC address for port %d, res = %d\n", port, res);
	return convert_error_ethAcc(res);
    }

    P_INFO("PHY %d MAC address is:\t", port);
    ixEthAccPortUnicastAddressShow(port);

    return 0;
}

/* set port MAC addr and update the dev struct if successfull */
int dev_set_mac_address(struct net_device *dev, void *addr)
{
    int res;
    priv_data_t *priv = dev->priv;
    IxEthAccMacAddr npeMacAddr;

    /* Set MAC addr in h/w */
    memcpy(&npeMacAddr.macAddress, &addr, IX_IEEE803_MAC_ADDRESS_SIZE);
    if ((res = ixEthAccPortUnicastMacAddressSet(priv->port_id, &npeMacAddr)))
    {
	P_ERROR("Failed to set MAC address for port %d, res = %d\n",
		priv->port_id, res);
	return convert_error_ethAcc(res);
    }

    /* update dev struct */
    memcpy(dev->dev_addr, &addr, IX_IEEE803_MAC_ADDRESS_SIZE);

    return 0;
}

/* Initialize device structs.
 * Resource allocation is deffered until do_dev_open
 */
static int __devinit ixp425_eth_probe(struct net_device *dev)
{
    int res;
    static int found_devices = 0;
    priv_data_t *priv;
    IxEthAccMacAddr mac_addr;

    TRACE;

    /* there are only 2 available ports */
    if (found_devices >= IX_ETH_ACC_NUMBER_OF_PORTS)
	return -ENODEV;

    SET_MODULE_OWNER(dev);

    /* set device name */
    strcpy(dev->name, found_devices ? DEVICE_NAME "1" : DEVICE_NAME "0");

    /* fill in dev struct callbacks */
    dev->open = do_dev_open;
    dev->stop = do_dev_stop;

    dev->hard_start_xmit = dev_hard_start_xmit;

    dev->watchdog_timeo = DEV_WATCHDOG_TIMEO;
    dev->tx_timeout = dev_tx_timeout;

    dev->do_ioctl = dev_do_ioctl;
    dev->get_stats = dev_get_stats;
    dev->set_multicast_list = dev_set_multicast_list;
    dev->flags |= IFF_MULTICAST;

    dev->set_mac_address = dev_set_mac_address;

    /* allocate and initialize priv struct */
    priv = dev->priv = kmalloc(sizeof(priv_data_t), GFP_KERNEL);
    if (dev->priv == NULL)
	return -ENOMEM;

    memset(dev->priv, 0, sizeof(priv_data_t));

    /* initialize RX pool */
    IX_MBUF_POOL_INIT_NO_ALLOC(&priv->rx_pool, priv->rx_pool_buf, NULL,
	RX_QUEUE_PREALLOC, 0, "IXP425 Ethernet driver Rx Pool");

    /* initialize TX pool */
    IX_MBUF_POOL_INIT_NO_ALLOC(&priv->tx_pool, priv->tx_pool_buf, NULL,
	TX_QUEUE_MAX_LEN, 0, "IXP425 Ethernet driver Tx Pool");

    priv->lock = SPIN_LOCK_UNLOCKED;
    priv->port_id = found_devices ? IX_ETH_PORT_2 : IX_ETH_PORT_1;

    /* initialise the MII register access mutex */
    priv->mediaCheckThreadComplete = (struct semaphore *)
	kmalloc(sizeof(struct semaphore), GFP_KERNEL);
    if (!priv->mediaCheckThreadComplete)
	return -ENOMEM;
    init_MUTEX(priv->mediaCheckThreadComplete);

    /* initialize ethernet device */
    ether_setup(dev);

    /* mac address */
    if ((res = ixEthAccPortUnicastMacAddressGet(priv->port_id, &mac_addr)))
	return convert_error_ethAcc(res);
    memcpy(dev->dev_addr, &mac_addr, IX_IEEE803_MAC_ADDRESS_SIZE);

    found_devices++;

    return 0;
}

/* Module initialization and cleanup */

#ifdef MODULE

static struct net_device ixp425_devices[IX_ETH_ACC_NUMBER_OF_PORTS];

int init_module(void)
{
    int res, i;
    struct net_device *dev;

    TRACE;

    P_INFO("Initializing " MODULE_NAME " driver " MODULE_VERSION "\n");

    TRACE;

    /* Do not initialise CSR core components if no_csr_init is set */
    if (no_csr_init) /* module parameter */
    {
	P_WARN("init_csr==0, no initialisation of CSR "
	       "core components performed\n");
    }
    else
    {
	if ((res = qmgr_init()))
	    return res;

	if ((res = ixNpeMhInitialize(IX_NPEMH_NPEINTERRUPTS_YES)))
	    return -1;
    }

    TRACE;

    for (i = 0; i < IX_ETH_ACC_NUMBER_OF_PORTS; i++)
    {
	if ((res = npe_start(i)))
	    return res;
    }

    if ((res = ethacc_init()))
	return res;

#ifndef CONFIG_IXP425_CSR_ETH_SKIP_PHY_SCAN
    if ((res = phy_init()))
	return res;
#endif

    if ((res = set_def_mac_addr(IX_ETH_PORT_1)))
	return res;

    if ((res = set_def_mac_addr(IX_ETH_PORT_2)))
	return res;

    /* Set scheduling discipline */
    ixEthAccTxSchedulingDisciplineSet(IX_ETH_PORT_1, FIFO_NO_PRIORITY);
    ixEthAccTxSchedulingDisciplineSet(IX_ETH_PORT_2, FIFO_NO_PRIORITY);

    for (i = 0; i < IX_ETH_ACC_NUMBER_OF_PORTS; i++)
    {
	dev = &ixp425_devices[i];

	dev->init = ixp425_eth_probe;

	if ((res = register_netdev(dev)))
	    P_ERROR("Failed to register netdev. res = %d\n", res);
    }

    maintenance_timer_set();

    /* initialise the MII register access mutex */
    miiAccessMutex = (struct semaphore *) kmalloc(sizeof(struct semaphore), GFP_KERNEL);
    if (!miiAccessMutex)
	return -ENOMEM;
    init_MUTEX(miiAccessMutex);

    return 0;
}

void cleanup_module(void)
{
    int i;

    TRACE;

    /* We can only get here when the module use count is 0,
     * so there's no need to stop devices.
     */

    maintenance_timer_clear();

    for (i = 0; i < IX_ETH_ACC_NUMBER_OF_PORTS; i++)
    {
	struct net_device *dev = &ixp425_devices[i];
	if (dev->priv != NULL)
	{
	    unregister_netdev(dev);
	    kfree(dev->priv);
	    dev->priv = NULL;
	}
    }
}

#endif
